# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyictacp', 'pyictacp.connection', 'pyictacp.packet', 'pyictacp.record']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'pyictacp',
    'version': '0.2.2',
    'description': 'ICT Protege Automation and Control Protocol wrapper',
    'long_description': None,
    'author': 'Thomas Hobson',
    'author_email': 'git@hexf.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
